#include <Wire.h>
#include <ESP8266WiFi.h>
#include <WiFiClient.h>
#include <PubSubClient.h>

#include "esp8266_hw.h"
#include "wifiCredentials.h"
#include "mqttCredentials.h"

#include <Adafruit_Sensor.h>
#include <Adafruit_BME280.h>
#include "SSD1306.h"

#define I2C_SDA 14
#define I2C_SCL 5

Adafruit_BME280 bme; // I2C
#define SEALEVELPRESSURE_HPA (1005)
//#define SEALEVELPRESSURE_HPA (1033.75)
unsigned long delayTime;
bool bBme280isConnected;

#include "SSD1306.h"
SSD1306  display(0x3c, I2C_SDA, I2C_SCL);

WiFiClient espClient;
PubSubClient client(espClient);

const byte button = 4;
boolean bWifiConnected = false;

void setup() {
  Serial.begin(115200);
  Wire.begin(I2C_SDA, I2C_SCL);
  vSetupBME280();
  vSetupOled();
}

// -----------------------------------------------------------------------------
// Generic {
// -----------------------------------------------------------------------------
void vDelayESP(unsigned long ulMilliseconds) {
  unsigned long ulPreviousMillis = millis();

  do {
    yield();
  } while (millis() - ulPreviousMillis <= ulMilliseconds);
}

void vGoingToDeepSleep(long int li) { // 5e6
  Serial.println("Going into deep sleep for 5 seconds");
  ESP.deepSleep(li); // 5e6 microseconds is 5 seconds //  According to the ESP8266 SDK, you can only sleep for 4,294,967,295 µs, which is about ~71 minutes.
}
// -----------------------------------------------------------------------------
// } Generic
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
// BME280 {
// -----------------------------------------------------------------------------
void vSetupBME280() {
  // initialize BME280 sensor
  bool status;
  status = bme.begin(0x76);
  Serial.println("");
  if (!status) {
    //Serial.println("Could not find a valid BME280 sensor, check wiring!");
    vBme280isNotConnected();
    bBme280isConnected = false;
  } else
    Serial.println("BME280 sensor is connected!");
  bBme280isConnected = true;
}

void vReadBME280(float* fTc, float* fP, float* fA, float* fRH) {
  *fTc = bme.readTemperature();
  *fP = bme.readPressure() / 100.0F;
  *fA = bme.readAltitude(SEALEVELPRESSURE_HPA);
  *fRH = bme.readHumidity();
}

void vPrintValues(float fT, float fP, float fA, float fH) {
  char sz[35];
  int nY = 0;

  display.clear();

  sprintf(sz, "T: %.2f C", fT);
  display.drawString(0, 0, sz);
  Serial.println(sz);
  sprintf(sz, "P: %.2f hPa", fP);
  display.drawString(0, nY += 15, sz);
  Serial.println(sz);
  sprintf(sz, "A: %.2f m", fA);
  display.drawString(0, nY += 15, sz);
  Serial.println(sz);
  sprintf(sz, "H: %.2f %%", fH);
  display.drawString(0, nY += 15, sz);
  Serial.println(sz);
  Serial.println();

  display.display();
}

void vBme280isNotConnected() {
  display.drawString(0, 15, "BME280");
  display.drawString(0, 30, "is NOT connected");
  Serial.println("BME280 is NOT connected");
}

void vPublicaDadesBme280(float fT, float fP, float fA, float fH) {
  char sz[33];
  
  sprintf(sz, "%f", fT);
  client.publish(TEMA_PUBLICA_TEMPERATURA, sz);
  //Serial.println(sz);
  sprintf(sz, "%f", fH);
  client.publish(TEMA_PUBLICA_HUMITAT, sz);
  //Serial.println(sz);
  sprintf(sz, "%f", fP);
  client.publish(TEMA_PUBLICA_PRESSIO, sz);
  //Serial.println(sz);
}

void vReadAndPrintBme280() {
  float fTc, fP, fA, fH;

  vReadBME280(&fTc, &fP, &fA, &fH);
  vPrintValues(fTc, fP, fA, fH);
  vPublicaDadesBme280(fTc, fP, fA, fH);
}
// -----------------------------------------------------------------------------
// } BME280
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
// OLED {
// -----------------------------------------------------------------------------
void vSetupOled() {
  display.init();
  display.flipScreenVertically();
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_LEFT);
}
// -----------------------------------------------------------------------------
// } OLED
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
// MQTT {
// -----------------------------------------------------------------------------
void callback(char* topic, byte* payload, unsigned int length) {
  String szRx = "", szTema(topic);

  Serial.print("Message arrived [");
  Serial.print(topic);
  Serial.print("] ");
  for (int i = 0; i < length; i++) {
    char receivedChar = (char)payload[i];
    szRx += receivedChar;
  }
  Serial.println(szRx);
  //if (szTema == TEMA_SUBSCRIPCIO_ESTAT_LEDS){}
}

void reconnect() {
  setup_wifi();
  client.setServer(mqtt_server, mqtt_port);
  client.setCallback(callback);
  // Loop until we're reconnected
  while (!client.connected()) {
    Serial.print("Attempting MQTT connection...");
    // Attempt to connect
    if (client.connect(NODE_UNIQUE_NAME, mqtt_user, mqtt_pwd)) {
      Serial.println("connected");
      // ... and subscribe to topic
      client.subscribe(TEMA_SUBSCRIPCIO_ESTAT_LEDS);
    } else {
      Serial.print("failed, rc=");
      Serial.print(client.state());
      Serial.println(" try again in 5 seconds");
      // Wait 5 seconds before retrying
      vDelayESP(5000);
    }
  }
}
// -----------------------------------------------------------------------------
// } MQTT
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
//  WiFi{
// -----------------------------------------------------------------------------
void vConnectToWiFi(const char* szSsid, const char* szPwd) {
  char ssid[MAX_STRING_SIZE],  pwd[MAX_STRING_SIZE];

  Serial.println("Connecting to WiFi network: " + String(szSsid) + ", pwd: " + String(szPwd));

  WiFi.begin(szSsid, szPwd);

  Serial.println("Waiting for WIFI connection...");
  while (WiFi.status() != WL_CONNECTED /*&& i++ < 20*/) {
    Serial.print(".");
    vDelayESP(500);
  }
  Serial.println();
}

boolean bIsListed(String szSSID, int *pNwO) {
  for (int i = 0; i < N_WIFIS ; i++) {
    if (String(stWiFi[i].szSSID) == szSSID) {
      *pNwO = i;
      return true;
    }
  }
  return false;
}

boolean bTryWifiConnection() {
  int n = WiFi.scanNetworks(), nWhichOne;

  //Serial.print("*");
  if (n == 0) {
    Serial.println("\nNo networks found");
    vDelayESP(1000);
  } else {
    for (int i = 0; i < n; ++i) {
      if (bIsListed(WiFi.SSID(i), &nWhichOne)) {
        vConnectToWiFi(stWiFi[nWhichOne].szSSID, stWiFi[nWhichOne].szPWD);
        return true;
      }
    }
  }
  return false;
}

void setup_wifi() {
  do {
    WiFi.mode(WIFI_STA);
    WiFi.disconnect();
    vDelayESP(100);
  } while (!bTryWifiConnection());
  IPAddress ip = WiFi.localIP();
  Serial.print("Connected to wifi at "); Serial.println(ip);
}

// -----------------------------------------------------------------------------
// }WiFi
// -----------------------------------------------------------------------------

void loop() {

  if (!client.connected()) {
    reconnect();
  }
  if (bBme280isConnected)
    vReadAndPrintBme280();

  client.loop();
  vDelayESP(1000);
  display.clear();
  display.display();
  //vDelayESP(5e3); // Time Oled is brighting
  vGoingToDeepSleep(5e6);
}

